package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State;
import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.Message;
import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.layout.Pane;
import javafx.stage.WindowEvent;

import java.io.IOException;

import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;

/**
 * Class Representing the Controller Element of the Window, also Contains the Elements that contacts the View Elements via
 * Listeners and Observable Objects.
 * To Contact the Model Elements needed it also holds references to message and the Connectionhandler.
 */
public class ChatWindowController {
    private ClientConnectionHandler connectionHandler;
    private ClientMessageList messages;
    private final WindowCloseHandler windowCloseHandler = new WindowCloseHandler();

    @FXML
    private Pane rootPane;
    @FXML
    private TextField serverAddressField;
    @FXML
    private TextField serverPortField;
    @FXML
    private TextField userNameField;
    @FXML
    private TextField messageField;
    @FXML
    private TextArea messageArea;
    @FXML
    private Button connectButton;
    @FXML
    private TextField filterValue;
    @FXML
    private Button sendButton;


    /**
     * Takes a message object and stores it used as Model and also starts message Listener via messageListener method.
     *
     * @param messages Object that will be set for use as Model
     */
    public void setMessages(ClientMessageList messages) {
        this.messages = messages;
        messageListener();
    }

    /**
     * Takes a Connection handler object and stores it used as Model and also starts a Listener for it.
     *
     * @param connectionHandler that will be set and used as Model.
     */
    public void setConnectionHandler(ClientConnectionHandler connectionHandler) {
        this.connectionHandler = connectionHandler;
        startConnectionHandlerListener();
        serverAddressField.setText(connectionHandler.getServerAddressProperty().get());
        serverPortField.setText(String.valueOf(connectionHandler.getServerPortProperty().get()));
    }

    /**
     * Method which closes the Application via use of the disconnect Method.
     */
    private void applicationClose() {
        disconnect();
    }

    /**
     * Method that handles the Connect Button and Initiates connect when connectionHandler is not Connected or
     * a disconnect when it is connected.
     */
    @FXML
    private void toggleConnection() {
        if (connectionHandler == null || connectionHandler.getStateProperty().get() != CONNECTED) {
            connect();
        } else {
            disconnect();
        }
    }

    /**
     * Initiates a connection by starting Connection Handler and telling, the Handler to start a connection.
     */
    private void connect() {
        try {
            messages.clear(); // clear message list
            startConnectionHandler();
            connectionHandler.connect();
        } catch (ChatProtocolException | IOException e) {
            addError("Error while starting Connection Handler and connect" + e);
        }
    }

    /**
     * Initiates disconnecting of the connectionHandler, also checks if connectionHandler is available.
     */
    private void disconnect() {
        if (connectionHandler == null) {
            addError("No connection handler");
            return;
        }
        try {
            connectionHandler.disconnect();
        } catch (ChatProtocolException e) {
            addError(e.getMessage());
        }
    }

    /**
     * Method which is used when the send button is pressed and handing over a message to the Connection Handler
     */
    @FXML
    private void message() {
        String messageString = messageField.getText().strip();
        try {
            if (connectionHandler == null) {
                addError("No connection handler");
            } else if (!connectionHandler.message(messageString)) {
                addError("Not a valid message format.");
            } else {
                messageField.clear();
            }
        } catch (ChatProtocolException e) {
            addError(e.getMessage());
        }
    }

    /**
     * Method which is used when a Filter is applied
     * Setting the Text in the message area after sending it through the filter.
     */
    @FXML
    private void applyFilter() {
        Platform.runLater(() -> this.messageArea.setText(messages.getFilteredMessages(filterValue.getText().strip())));
    }

    /**
     * Starts the ConnectionHandler setting the username and Checking if the name follows the valid format of no spaces.
     *
     * @throws IOException for error that may occur during initialization of connectionHandler.
     */
    private void startConnectionHandler() throws IOException {
        String userName = userNameField.getText();
        if (!userName.contains(" ")) {
            String serverAddress = serverAddressField.getText();
            int serverPort = Integer.parseInt(serverPortField.getText());
            connectionHandler.initialize(serverAddress, serverPort, userName);
            new Thread(connectionHandler).start();

            rootPane.getScene().getWindow().addEventHandler(WindowEvent.WINDOW_CLOSE_REQUEST, windowCloseHandler);
        } else {
            addError("It is not allowed to have spaces in username!");
        }
    }

    /**
     * Sets the state shown according to the state the method receives, if state indicates disconnected it will also inform the
     * Connection Handler and tell it to stop Receiving more messages.
     *
     * @param newState is the state that it should be set to.
     */
    public void stateChanged(State newState) {
        // update UI (need to be run in UI thread: see Platform.runLater())
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                connectButton.setText((newState == CONNECTED || newState == CONFIRM_DISCONNECT) ? "Disconnect" : "Connect");
            }
        });
        if (newState == DISCONNECTED) {
            connectionHandler.stopReceiving();
        }
    }

    /**
     * Sets displayed username according to the String provided.
     *
     * @param userName provided String that is set as name.
     */
    public void setUserName(String userName) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                userNameField.setText(userName);
            }
        });
    }

    /**
     * Sets displayed Server Address.
     *
     * @param serverAddress provided String that is set as server address.
     */
    public void setServerAddress(String serverAddress) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverAddressField.setText(serverAddress);
            }
        });
    }

    /**
     * Sets displayed Server port.
     *
     * @param serverPort provided String that is set as server port.
     */
    public void setServerPort(int serverPort) {
        Platform.runLater(new Runnable() {
            @Override
            public void run() {
                serverPortField.setText(Integer.toString(serverPort));
            }
        });
    }

    /**
     * Method which adds an incoming String as an Error Message.
     *
     * @param message String to be added as Error
     */
    public void addError(String message) {
        messages.addMessage(new Message(ConnectionHandler.DATA_TYPE.DATA_TYPE_ERROR, null, null, message));
    }

    /**
     * Nested Class in charge of Closing the window
     */
    class WindowCloseHandler implements EventHandler<WindowEvent> {

        /**
         *
         * @param event the event which occurred when Windows is closed
         */
        public void handle(WindowEvent event) {
            applicationClose();
        }
    }

    /**
     * Starts several new Listener for Connection Handler changes by using several observable properties.
     */
    public void startConnectionHandlerListener() {
        connectionHandler.getStateProperty().addListener(new ChangeListener<State>() {
            @Override
            public void changed(ObservableValue<? extends State> observable, State oldValue, State newValue) {
                stateChanged(newValue);
            }
        });

        connectionHandler.getUserNameProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setUserName(newValue);
            }
        });

        connectionHandler.getServerAddressProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                setServerAddress(newValue);
            }
        });

        connectionHandler.getServerPortProperty().addListener(new ChangeListener<Number>() {
            @Override
            public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue) {
                setServerPort(newValue.intValue());
            }
        });
    }

    /**
     * Starts a new Listener for messages by using the observable Boolean.
     */
    private void messageListener() {
        messages.getChangedProperty().addListener(new ChangeListener<Boolean>() {
            @Override
            public void changed(ObservableValue<? extends Boolean> observable, Boolean oldValue, Boolean newValue) {
                Platform.runLater(() -> messageArea.setText(messages.getFilteredMessages(filterValue.getText().strip())));
            }
        });
    }

}
