package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.Message;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Scanner;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;

/**
 * Client Connection Handler Class is used for the connection of the Client to a server
 * it is used part as a Model storing properties like userName, state, serverAddress and Port.
 * Also holds methods to
 */
public class ClientConnectionHandler extends ConnectionHandler implements Runnable {

    private final Pattern messagePattern = Pattern.compile("^(?:@(\\S*))?\\s*(.*)$");

    private final SimpleStringProperty userName;
    private final SimpleObjectProperty<State> state;
    private final ClientMessageList messages;
    private final SimpleStringProperty serverAddress;
    private final SimpleIntegerProperty serverPort;

    /**
     * Constructor initializes ConnectionHandler by Setting default values for the fields and stores the given messages itself.
     *
     * @param messages
     */
    public ClientConnectionHandler(ClientMessageList messages) {
        super();
        this.messages = messages;
        state = new SimpleObjectProperty<>(State.NEW);
        serverAddress = new SimpleStringProperty(NetworkHandler.DEFAULT_ADDRESS.getCanonicalHostName());
        serverPort = new SimpleIntegerProperty(NetworkHandler.DEFAULT_PORT);
        this.userName = new SimpleStringProperty(null);
    }

    /**
     * Called to initialize the ClientConnectionHandler when trying to start a connection
     *
     * @param serverAddress to connect to
     * @param serverPort to connect to
     * @param userName to connect as
     * @throws IOException if connection to Server not possible
     */
    public void initialize(String serverAddress, int serverPort, String userName) throws IOException {
        state.set(NEW);
        this.serverAddress.set(serverAddress);
        this.serverPort.set(serverPort);
        setConnection(NetworkHandler.openConnection(serverAddress, serverPort));
        this.userName.set((userName == null || userName.isBlank()) ? USER_NONE : userName);
    }

    /**
     * Observable getter Method for the stored serverAddress
     *
     * @return the stored serverAddress
     */
    public SimpleStringProperty getServerAddressProperty() {
        return serverAddress;
    }

    /**
     * Observable getter Method for the stored serverPort
     *
     * @return the stored serverPort
     */
    public SimpleIntegerProperty getServerPortProperty() {
        return serverPort;
    }

    /**
     * Observable getter Method for the stored state
     *
     * @return the stored state
     */
    public SimpleObjectProperty<State> getStateProperty() {
        return this.state;
    }

    /**
     * Observable getter Method for the stored userName
     *
     * @return the stored userName
     */
    public SimpleStringProperty getUserNameProperty() {
        return userName;
    }

    /**
     * Method which sets a new State.
     *
     * @param newState the state that will be set
     */
    public void setState(State newState) {
        state.set(newState);
    }

    /**
     * Standard run method which will directly start the startReceiving method.
     */
    public void run() {
        startReceiving();
    }

    /**
     * Method that is started by the run method, starts a connection handler.
     * Figures out if connection is available if not determines the error cause and gives an error accordingly.
     */
    private void startReceiving() {
        System.out.println("Starting Connection Handler");
        try {
            System.out.println("Start receiving data...");
            while (getConnection().isAvailable()) {
                Message data = getConnection().receive();
                processData(data);
            }
            System.out.println("Stopped receiving data");
        } catch (SocketException e) {
            System.out.println("Connection terminated locally");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch (EOFException e) {
            System.out.println("Connection terminated by remote");
            this.setState(DISCONNECTED);
            System.err.println("Unregistered because connection terminated" + e.getMessage());
        } catch (IOException e) {
            System.err.println("Communication error" + e);
        } catch (ClassNotFoundException e) {
            System.err.println("Received object of unknown type" + e.getMessage());
        }
        System.out.println("Stopped Connection Handler");
    }

    /**
     * Method which is used to stop receiving data, gets the current connection and closes it.
     */
    public void stopReceiving() {
        System.out.println("Closing Connection Handler to Server");
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler to Server");
    }

    /**
     * Method which processes data and determines its type then uses the corresponding method to process it.
     *
     * @param data that is received in a form of a String and then used depending on its determined cause.
     */
    private void processData(Message data) {
        // dispatch operation based on type parameter
        if (data.getType() == DATA_TYPE.DATA_TYPE_CONNECT) {
            System.err.println("Illegal connect request from server");
        } else if (data.getType() == DATA_TYPE.DATA_TYPE_CONFIRM) {
            caseConfirm(data);
        } else if (data.getType() == DATA_TYPE.DATA_TYPE_DISCONNECT) {
            caseDisconnect(data);
        } else if (data.getType() == DATA_TYPE.DATA_TYPE_MESSAGE) {
            caseMessage(data);
        } else if (data.getType() == DATA_TYPE.DATA_TYPE_ERROR) {
            caseError(data);
        } else {
            System.out.println("Unknown data type received: " + data.getType());
        }
    }

    private void caseConfirm(Message data) {
        if (state.get() == CONFIRM_CONNECT) {
            this.userName.set(data.getReceiver());
            this.serverPort.set(getConnection().getRemotePort());
            this.serverAddress.set(getConnection().getRemoteHost());
            messages.addMessage(data);
            System.out.println("CONFIRM: " + data.getText());
            this.setState(CONNECTED);
        } else if (state.get() == CONFIRM_DISCONNECT) {
            messages.addMessage(data);
            System.out.println("CONFIRM: " + data.getText());
            this.setState(DISCONNECTED);
        } else {
            System.err.println("Got unexpected confirm message: " + data.getText());
        }
    }

    /**
     * Initiates the disconnect sequence and sends the message with all its info.
     *
     * @param data Data which has been transmitted
     */
    private void caseDisconnect(Message data) {
        if (state.get() == DISCONNECTED) {
            System.out.println("DISCONNECT: Already in disconnected: " + data.getText());
            return;
        }
        messages.addMessage(data);
        System.out.println("DISCONNECT: " + data.getText());
        this.setState(DISCONNECTED);
    }

    /**
     * Initiates the procedure to send a new message and sends one as such.
     * @param data Data which has been transmitted
     */
    private void caseMessage(Message data) {
        if (state.get() != CONNECTED) {
            System.out.println("MESSAGE: Illegal state " + state + " for message: " + data.getText());
            return;
        }
        messages.addMessage(data);
        System.out.println("MESSAGE: From " + data.getSender() + " to " + data.getReceiver() + ": " + data.getText());
    }

    /**
     * Stores the message as an error message and displays it as such as well.
     *
     * @param data Data which has been transmitted
     */
    private void caseError(Message data) {
        messages.addMessage(data);
        System.out.println("ERROR: " + data.getText());
    }

    /**
     * Send connect attempt to Server
     *
     * @throws ChatProtocolException Error that is thrown if the state is not set to NEW
     */
    public void connect() throws ChatProtocolException {
        if (state.get() != NEW) throw new ChatProtocolException("Illegal state for connect: " + state);
        this.sendData(userName.get(), USER_NONE, getDataTypeConnect(), null);
        this.setState(CONFIRM_CONNECT);
    }

    /**
     * Send Disconnect attempt to server
     *
     * @throws ChatProtocolException Error tha si thrown if state is invalid
     */
    public void disconnect() throws ChatProtocolException {
        if (state.get() != NEW && state.get() != CONNECTED)
            throw new ChatProtocolException("Illegal state for disconnect: " + state);
        this.sendData(userName.get(), USER_NONE, getDataTypeDisconnect(), null);
        this.setState(CONFIRM_DISCONNECT);
    }

    /**
     * Sends message to Server
     *
     * @param messageString The message the user wants to send
     * @return true if message is valid else false
     * @throws ChatProtocolException if illegal connection state
     */
    public boolean message(String messageString) throws ChatProtocolException {
        if (state.get() != CONNECTED) throw new ChatProtocolException("Illegal state for message: " + state);

        Matcher matcher = messagePattern.matcher(messageString);
        if (matcher.find()) {
            String receiver = matcher.group(1);
            String message = matcher.group(2);
            if (message.length() < 1) {
                return false;
            }
            if (receiver == null || receiver.isBlank()) receiver = ClientConnectionHandler.USER_ALL;
            this.sendData(userName.get(), receiver, getDataTypeMessage(), message);
            return true;
        } else {
            return false;
        }
    }
}
