package ch.zhaw.pm2.multichat.client;

import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;
import ch.zhaw.pm2.multichat.protocol.Message;
import javafx.beans.property.SimpleBooleanProperty;

import java.util.ArrayList;
import java.util.List;

/**
 * Class that is used to store the messages in an ArrayList
 * And also as a Model that informs the controller via Listeners that can be initialized
 */
public class ClientMessageList {
    private List<Message> messages = new ArrayList<>();
    private final SimpleBooleanProperty changed = new SimpleBooleanProperty(false);

    /**
     * Adds a new message to ArrayList and also informs Listener.
     * @param message that should be added
     */
    public void addMessage(Message message) {
        messages.add(message);
        changed.set(!changed.get());
    }

    /**
     * Applies a given filter over all messages and returns the result as a new String.
     * Method is also in charge of applying the correct prefix to the message according to its type.
     *
     * @param filter is the applied filter on all the messages that are stored.
     * @return String that matches the given filter.
     */
    public String getFilteredMessages(String filter) {
        StringBuilder result = new StringBuilder();
        boolean showAll = filter == null || filter.isBlank();
        for (Message message : messages) {
            if (showAll || message.matchesFilter(filter)) {
                switch (message.getType()) {
                    case DATA_TYPE_MESSAGE ->
                            result.append(String.format("[%s -> %s] %s\n", message.getSender(), message.getReceiver(), message.getText()));
                    case DATA_TYPE_ERROR -> result.append(String.format("[ERROR] %s\n", message.getText()));
                    case DATA_TYPE_CONFIRM, DATA_TYPE_DISCONNECT, DATA_TYPE_CONNECT -> result.append(String.format("[INFO] %s\n", message.getText()));
                    default ->
                            result.append(String.format("[ERROR] %s\n", "Unexpected message type: " + message.getType()));
                }
            }
        }
        return result.toString();
    }

    /**
     * Overwrites the Arraylist of messages, clearing it, also informs all Listeners.
     */
    public void clear() {
        messages = new ArrayList<>();
        changed.set(!changed.get());
    }

    /**
     * Getter Method to check the current value of SimpleBooleanProperty changed.
     *
     * @return the current value of changed
     */
    public SimpleBooleanProperty getChangedProperty() {
        return changed;
    }

}
