package ch.zhaw.pm2.multichat.protocol;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Scanner;

/**
 * This abstract class is the superclass for ClientConnectionHandler and ServerConnectionHandler
 * It offers the DATA_TYPE Strings and a {@link State} enum for all valid connection states.
 * Shared methods are implemented in this class as well {@link ConnectionHandler#sendData(String, String, DATA_TYPE, String)} {@link ConnectionHandler#processData(Scanner, StringBuilder, StringBuilder, StringBuilder, StringBuilder)}
 */
public abstract class ConnectionHandler {
    private NetworkHandler.NetworkConnection<String> connection;

    public static final String USER_NONE = "";
    public static final String USER_ALL = "*";

    // State of the connection
    public enum State {
        NEW, CONFIRM_CONNECT, CONNECTED, CONFIRM_DISCONNECT, DISCONNECTED, ERROR;
    }

    public enum DATA_TYPE {
        DATA_TYPE_CONNECT, DATA_TYPE_CONFIRM, DATA_TYPE_DISCONNECT,DATA_TYPE_MESSAGE,DATA_TYPE_ERROR
    }

    /**
     * @return {@link DATA_TYPE}
     */
    public static DATA_TYPE getDataTypeConnect() {
        return DATA_TYPE.DATA_TYPE_CONNECT;
    }

    /**
     * @return {@link DATA_TYPE}
     */
    public static DATA_TYPE getDataTypeConfirm() {
        return DATA_TYPE.DATA_TYPE_CONFIRM;
    }

    /**
     * @return {@link DATA_TYPE}
     */
    public static DATA_TYPE getDataTypeDisconnect() {
        return DATA_TYPE.DATA_TYPE_DISCONNECT;
    }

    /**
     * @return {@link DATA_TYPE}
     */
    public static DATA_TYPE getDataTypeMessage() {
        return DATA_TYPE.DATA_TYPE_MESSAGE;
    }

    /**
     * @return {@link DATA_TYPE}
     */
    public static DATA_TYPE getDataTypeError() {
        return DATA_TYPE.DATA_TYPE_ERROR;
    }

    /**
     * @return {@link NetworkHandler.NetworkConnection}
     */
    public NetworkHandler.NetworkConnection<String> getConnection() {
        return connection;
    }

    /**
     * This method sets the NetworkConnection used for the server <-> client connection
     *
     * @param connection NetworkConnection used for the server <-> client connection
     */
    protected void setConnection(NetworkHandler.NetworkConnection<String> connection) {
        this.connection = connection;
    }


    /**
     * This method gets called to send data via the socket defined in the {@link NetworkHandler.NetworkConnection}
     *
     * @param sender   of the data
     * @param receiver of the data
     * @param type     of the data
     * @param payload  of the data
     */
    protected void sendData(String sender, String receiver, DATA_TYPE type, String payload) {
        if (connection.isAvailable()) {
            try {
                connection.send(new Message(type,sender,receiver,payload));
            } catch (SocketException e) {
                System.err.println("Connection closed: " + e.getMessage());
            } catch (EOFException e) {
                System.out.println("Connection terminated by remote");
            } catch (IOException e) {
                System.err.println("Communication error: " + e.getMessage());
            }
        }
    }
}
