package ch.zhaw.pm2.multichat.server;

import ch.zhaw.pm2.multichat.protocol.ChatProtocolException;
import ch.zhaw.pm2.multichat.protocol.ConnectionHandler;

import static ch.zhaw.pm2.multichat.protocol.ConnectionHandler.State.*;

import ch.zhaw.pm2.multichat.protocol.Message;
import ch.zhaw.pm2.multichat.protocol.NetworkHandler;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketException;
import java.util.Map;
import java.util.Objects;
import java.util.Scanner;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;


/**
 * This class represents the connection between the server and a client and offers the serverside logic.
 * The ServerConnectionHandler receives data send from the client as well as sends data to the client.
 * <p>
 * The ServeConnectionHandler offers following functionality:
 * <p>
 * Evaluating connection attempts from a client by:
 * 1. Checks if Username is valid (Not used)
 * 2. Saves Username in {@link ServerConnectionHandler#userName}
 * 3. Saves the connection in the {@link ServerConnectionHandler#connectionRegistry}
 * <p>
 * Processes disconnections from a client by:
 * 1. Removing the connection from the {@link ServerConnectionHandler#connectionRegistry}
 * 2. Terminates the socket by calling {@link NetworkHandler.NetworkConnection#close()}
 * <p>
 * Process Messages send from a client by:
 * 1. Evaluating the receiver by differentiating between broadcast or unicast.
 * 2. Sending the message accordingly.
 * <p>
 * To use this class, start a new instance and start it in a thread.
 * To constructor needs following parameter:
 * 1. {@link ch.zhaw.pm2.multichat.protocol.NetworkHandler.NetworkConnection} representing the socket connection between server and client
 * 2. {@link Map<String,ServerConnectionHandler>} registry to check for all active connections
 * 3. {@link ReentrantLock @link Condition} to lock server thread to evaluate connection.
 */
public class ServerConnectionHandler extends ConnectionHandler implements Runnable {
    private static final AtomicInteger connectionCounter = new AtomicInteger(0);
    private final int connectionId = connectionCounter.incrementAndGet();
    private final Map<String, ServerConnectionHandler> connectionRegistry;

    private String userName = "Anonymous-" + connectionId;
    private State state = NEW;

    /**
     * Called when runnable gets started in a thread.
     */
    @Override
    public void run() {
        startReceiving();
    }

    /**
     * Constructor to initialize the connection
     *
     * @param connection   representing the socket connection between server and client
     * @param registry     map containing all active connections between server and clients


     */
    public ServerConnectionHandler(NetworkHandler.NetworkConnection<String> connection,
                                   Map<String, ServerConnectionHandler> registry) {
        super();
        setConnection(connection);
        Objects.requireNonNull(connection, "Connection must not be null");
        Objects.requireNonNull(registry, "Registry must not be null");
        this.connectionRegistry = registry;
    }

    /**
     * These methods runs in a while-loop as long as the socket between server and client is available
     * and the connection State is not ERROR.
     */
    private void startReceiving() {
        System.out.println("Starting Connection Handler for new User");
        try {
            System.out.println("Start receiving data...");
            while (getConnection().isAvailable() && !(state == ERROR)) {
                Message data = getConnection().receive();
                processData(data);
            }
            System.out.println("Stopped receiving data");
        } catch (SocketException e) {
            System.out.println("Connection terminated locally");
            connectionRegistry.remove(userName);
            System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
        } catch (EOFException e) {
            System.out.println("Connection terminated by remote");
            connectionRegistry.remove(userName);
            System.out.println("Unregistered because client connection terminated: " + userName + " " + e.getMessage());
        } catch (IOException e) {
            System.err.println("Communication error: " + e);
        } catch (ClassNotFoundException e) {
            System.err.println("Received object of unknown type: " + e.getMessage());
        }
        if (state == ERROR) {
            System.out.println("Stopping Connection Handler for Rejected Client");
        } else {
            System.out.println("Stopping Connection Handler for " + userName);
        }

    }

    /**
     * This method will call {@link NetworkHandler.NetworkConnection#close()} to close the Socket.
     */
    private void stopReceiving() {
        System.out.println("Closing Connection Handler for " + userName);
        try {
            System.out.println("Stop receiving data...");
            getConnection().close();
            System.out.println("Stopped receiving data.");
        } catch (IOException e) {
            System.err.println("Failed to close connection." + e.getMessage());
        }
        System.out.println("Closed Connection Handler for " + userName);
    }

    /**
     * This method gets called when socket receives data. The method checks for the data type and reacts accordingly
     * If data type:
     * 1. Connect => checks if username is valid. if valid sends response to client with confirmation.
     * If username not valid quits connection by changing status to ERROR.
     * 2. Confirm => Server should not receive this kind of message. STDOUT informs about it.
     * 3. Disconnect => Disconnects connection by removing connection from registry and calling method to terminate socket.
     * 4. Message => Checks if broadcast or unicast. Sends data accordingly
     * 5. Error => STDERR message
     *
     * @param data received by the server
     */
    private void processData(Message data) {
        try {
            // dispatch operation based on type parameter
            if (data.getType() == DATA_TYPE.DATA_TYPE_CONNECT) {
                caseConnect(data);
            } else if (data.getType() == DATA_TYPE.DATA_TYPE_CONFIRM) {
                System.out.println("Not expecting to receive a CONFIRM request from client");
            } else if (data.getType() == DATA_TYPE.DATA_TYPE_DISCONNECT) {
                caseDisconnect();
            } else if (data.getType() == DATA_TYPE.DATA_TYPE_MESSAGE) {
                caseMessage(data);
            } else if (data.getType() == DATA_TYPE.DATA_TYPE_ERROR) {
                System.err.println("Received error from client (" + data.getSender() + "): " + data.getText());
            } else {
                System.err.println("Unknown data type received: " + data.getType());

            }
        } catch (ChatProtocolException e) {
            System.out.println("Error while processing data " + e.getMessage());
            sendData(USER_NONE, userName, getDataTypeError(), e.getMessage());
        }
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(Message)}
     * Checks if username is valid. if valid sends response to client with confirmation.
     *
     * @param data sent
     * @throws ChatProtocolException if username not valid
     */
    private void caseConnect(Message data) throws ChatProtocolException {
        if (this.state != NEW) throw new ChatProtocolException("Illegal state for connect request: " + state);
        if (data.getSender().isBlank()) data.setSender(this.userName);
        //if username not valid
        if (connectionRegistry.containsKey(data.getSender())) {
            state = ERROR;
            System.out.println(String.format("Connecting failed for new Client with IP:Port <%s:%d>.\nReason: Name already taken.",
                    getConnection().getRemoteHost(),
                    getConnection().getRemotePort()));
            throw new ChatProtocolException("User name already taken: " + data.getSender());
        }
        //if username valid
        this.userName = data.getSender();
        connectionRegistry.put(userName, this);
        sendData(USER_NONE, userName, getDataTypeConfirm(), "Registration successful for " + userName);
        this.state = CONNECTED;
        System.out.println(String.format("Connected new Client %s with IP:Port <%s:%d>",
                userName,
                getConnection().getRemoteHost(),
                getConnection().getRemotePort()));
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(Message)}
     * Disconnects connection by removing connection from registry and calling method {@link ServerConnectionHandler#stopReceiving()} to terminate socket.
     *
     * @throws ChatProtocolException if state already DISCONNECTED.
     */
    private void caseDisconnect() throws ChatProtocolException {
        if (state == DISCONNECTED)
            throw new ChatProtocolException("Illegal state for disconnect request: " + state);
        if (state == CONNECTED) {
            connectionRegistry.remove(this.userName);
        }
        sendData(USER_NONE, userName, getDataTypeConfirm(), "Confirm disconnect of " + userName);
        this.state = DISCONNECTED;
        this.stopReceiving();
    }

    /**
     * This method is called by method {@link ServerConnectionHandler#processData(Message)}
     * Checks if broadcast or unicast. Sends data accordingly
     *
     * @param data to transmit
     * @throws ChatProtocolException if state not equal to CONNECT
     */
    private void caseMessage(Message data) throws ChatProtocolException {
        if (state != CONNECTED) throw new ChatProtocolException("Illegal state for message request: " + state);
        if (USER_ALL.equals(data.getReceiver())) {
            for (ServerConnectionHandler handler : connectionRegistry.values()) {
                handler.sendData(data.getSender(), data.getReceiver(), data.getType(), data.getText());
            }
        } else {
            ServerConnectionHandler handler = connectionRegistry.get(data.getReceiver());
            if (handler != null) {
                handler.sendData(data.getSender(), data.getReceiver(), data.getType(), data.getText());
                if (!data.getReceiver().equals(data.getSender())) {
                    sendData(data.getSender(), data.getReceiver(), data.getType(), data.getText()); //send message to sender if it's a direct message and sender is not receiver.
                }
            } else {
                this.sendData(USER_NONE, userName, getDataTypeError(), "Unknown User: " + data.getReceiver());
            }
        }
    }
}
